SpellWatch.Menu = { };
SpellWatch.Menu.list = { };
SpellWatch.Menu.offset = 0;
SpellWatch.Menu.selection = nil;

-- Get the selection string and table
function SpellWatch.Menu:GetSelection()

	if not self.selection then return nil; end
	
	return self.list[self.selection], SpellWatch.db.profile.timers[self.list[self.selection]];

end

-- New selection
function SpellWatch.Menu:New(name)

	if not SpellWatch.db.profile.timers[name] then
		SpellWatch.db.profile.timers[name] = {spell = "", label = "", duration = 0, type = SPELLWATCH_TYPE_SPELL, color = "808080"};
	end

	self:ListBuild();
	self:SetSelection(name);
	self:ListUpdate();

end

-- Delete selection
function SpellWatch.Menu:Delete(name)

	if not name then return; end

	SpellWatch.db.profile.timers[name] = nil;
	self:ListBuild();
	self:ListUpdate();

end

-- Load a given variable from the currently selected timer config
function SpellWatch.Menu:LoadVar(var)

	if not var then return nil; end

	local _, config = self:GetSelection();

	if not config then return nil; end

	return config[var];

end

-- Save a given variable to the currently selected timer config
function SpellWatch.Menu:SaveVar(var, val)

	if not var then return nil; end

	local _, config = self:GetSelection();

	if not config then return nil; end

	config[var] = val;

end

-- Set the selection to a specified nam
function SpellWatch.Menu:SetSelection(name)

	self.selection = nil;
	local i,n;

	-- match the name
	for i,n in ipairs(self.list) do if n == name then self.selection = i; break; end end

	self:ListUpdate();
	self:LoadConfig();

end

-- Set the selection based on a button click
function SpellWatch.Menu:ListEntryOnClick(id)

	if not id then return; end

	self:SetSelection(self.list[self.offset + id]);

end

-- Build the list of timers
function SpellWatch.Menu:ListBuild()

	local name, config;

	-- Clear the list
	for name in pairs(self.list) do self.list[name] = nil; end

	-- Reinitialize the list
	for name in pairs(SpellWatch.db.profile.timers) do table.insert(self.list, name); end

	-- Sort alphabetically
	table.sort(self.list);
	
	self.selection = self.list[1] and 1;

end

-- Update the list of timers
function SpellWatch.Menu:ListUpdate()

	local timer;

	FauxScrollFrame_Update(SpellWatchMenuFrameScrollBar, table.getn(self.list), 15, 16);

	self.offset = FauxScrollFrame_GetOffset(SpellWatchMenuFrameScrollBar);

	for timer = 1, 15 do

		local button = getglobal("SpellWatchMenuFrameListEntry"..timer);
		local index = self.offset + timer;

		if ( index <= table.getn(self.list) ) then
			button:SetText(self.list[index]);
			button:Show();
			if self.selection and self.selection == index then
				button:LockHighlight();
			else
				button:UnlockHighlight();
			end
		else
			button:Hide();
		end

	end

end

-- Update the entire menu with the currently selected config
function SpellWatch.Menu:LoadConfig()

	local loadText = function ( frame ) getglobal(frame:GetName().."EditBox"):SetText(frame.load() or ""); end
	local loadCheck = function ( frame ) getglobal(frame:GetName().."CheckBox"):SetChecked(frame.load()); end

	loadText(SpellWatchMenuFrameTitle);
	loadText(SpellWatchMenuFrameSpell);
	loadText(SpellWatchMenuFrameLabel);
	loadText(SpellWatchMenuFrameDuration);
	SpellWatchMenuFrameType.load();
	loadText(SpellWatchMenuFrameGroup);
	loadText(SpellWatchMenuFrameColor);
	SpellWatchMenuFrameColorPicker:GetNormalTexture():SetTexture(SpellWatch.h2f(self:LoadVar("color")));
	SpellWatchMenuFrameIcon:SetNormalTexture(self:LoadVar("icon") or "Interface\\Icons\\INV_Misc_QuestionMark");
	SpellWatchMenuFrameIcon:SetPushedTexture(self:LoadVar("icon") or "Interface\\Icons\\INV_Misc_QuestionMark");
	loadCheck(SpellWatchMenuFrameTargeted);
	loadCheck(SpellWatchMenuFrameUnique);
	loadCheck(SpellWatchMenuFrameAnnounce);
	--loadCheck(SpellWatchMenuFrameInvertBar);
	--loadCheck(SpellWatchMenuFrameInvertTimer);

end
